package main

import "github.com/veandco/go-sdl2/sdl"

/*
#cgo CFLAGS: -I/home/nick/Downloads/include
#cgo LDFLAGS: -L/home/nick/.local/lib -lprojectM-4
#include "projectM-4/projectM.h"
#include <stdlib.h>
*/
import "C"

import (
	"encoding/binary"
	"errors"
	"io"
	"os"
	"unsafe"
)

const bufSize uint = 2048

type milkDropWindow struct {
	window  *sdl.Window
	context sdl.GLContext
	handle  C.projectm_handle
}

func newMilkDropWindow(width, height int32) (*milkDropWindow, error) {
	var m milkDropWindow
	var err error

	m.window, err = sdl.CreateWindow("milkbucket", sdl.WINDOWPOS_UNDEFINED, sdl.WINDOWPOS_UNDEFINED, width, height, sdl.WINDOW_OPENGL|sdl.WINDOW_RESIZABLE)
	if err != nil {
		return &m, err
	}

	m.context, err = m.window.GLCreateContext()
	if err != nil {
		return &m, err
	}

	m.handle = C.projectm_create()
	if m.handle == nil {
		return &m, errors.New("error creating projectM instance")
	}

	C.projectm_set_window_size(m.handle, C.ulong(width), C.ulong(height))

	return &m, nil
}

func (m *milkDropWindow) destroy() {
	C.projectm_destroy(m.handle)
	sdl.GLDeleteContext(m.context)
	m.window.Destroy()
}

func (m *milkDropWindow) loadPreset(preset string) {
	cPreset := C.CString("/usr/share/projectM/presets/presets_stock/Zylot - Wisps.milk")
	defer C.free(unsafe.Pointer(cPreset))
	C.projectm_load_preset_file(m.handle, cPreset, false)
}

func (m *milkDropWindow) render() (bool, error) {
	var audioData [bufSize]int16

	err := binary.Read(os.Stdin, binary.LittleEndian, &audioData)
	if err == io.ErrUnexpectedEOF {
		return false, nil
	} else if err != nil {
		return true, err
	}

	C.projectm_pcm_add_int16(m.handle, (*C.int16_t)(&audioData[0]), C.uint(bufSize), C.PROJECTM_STEREO)

	C.projectm_opengl_render_frame(m.handle)
	m.window.GLSwap()

	return true, nil
}

func handleWindowEvent(event *sdl.WindowEvent, handle C.projectm_handle) {
	switch event.Event {
	case sdl.WINDOWEVENT_RESIZED:
		w, h := event.Data1, event.Data2
		C.projectm_set_window_size(handle, C.ulong(w), C.ulong(h))
	}
}

func handleEvent(event sdl.Event, handle C.projectm_handle) bool {
	switch event.(type) {
	case *sdl.QuitEvent:
		return false
	case *sdl.WindowEvent:
		event := event.(*sdl.WindowEvent)
		handleWindowEvent(event, handle)
	}

	return true
}

func main() {
	err := sdl.Init(sdl.INIT_VIDEO)
	if err != nil {
		panic(err)
	}
	defer sdl.Quit()

	m, err := newMilkDropWindow(800, 600)
	if err != nil {
		panic(err)
	}
	defer m.destroy()

	m.loadPreset("/usr/share/projectM/presets/presets_stock/Zylot - Wisps.milk")

	running := true
	for running {
		for event := sdl.PollEvent(); event != nil; event = sdl.PollEvent() {
			running = handleEvent(event, m.handle)
			if !running {
				break
			}
		}

		if !running {
			break
		}

		running, err = m.render()
		if err != nil {
			panic(err)
		}
	}
}
